SystemOrganization addCategory: #'DnsClient-Core'!

Object subclass: #DnsClient
	instanceVariableNames: 'mutex servers cache nextUpdate'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'DnsClient-Core'!

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:29'!
addressForName: hostName
	"Return an address for the given host name"

	^self default addressForName: hostName! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:29'!
addressForName: hostName timeout: seconds
	"Return an address for the given host name"

	^self default addressForName: hostName timeout: seconds! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:29'!
addressesForName: hostName
	"Return all the (IPv4) addresses for the given hostName"

	^self default addressesForName: hostName! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:28'!
addressesForName: hostName timeout: seconds
	"Return all the (IPv4) addresses for the given hostName"

	^self default addressesForName: hostName timeout: seconds! !

!DnsClient class methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:26'!
default
	"Answer the default DnsClient to use if no other is specified"

	^Default ifNil:[Default := self new].! !

!DnsClient class methodsFor: 'class initialization' stamp: 'ar 7/26/2010 22:44'!
initialize
	"DnsClient initialize"

	Smalltalk addToShutDownList: self.! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:28'!
mxRecordsFor: hostName
	"Return the MX records for the given host name. NOT cached."

	^self default mxRecordsFor: hostName! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:28'!
mxRecordsFor: hostName timeout: seconds
	"Return the MX records for the given host name. NOT cached."

	^self default mxRecordsFor: hostName timeout: seconds! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:28'!
nameForAddress: address
	"Perform a reverse DNS lookup for the given address"

	^self default nameForAddress: address! !

!DnsClient class methodsFor: 'resolving' stamp: 'ar 7/26/2010 22:28'!
nameForAddress: address timeout: seconds
	"Perform a reverse DNS lookup for the given address"

	^self default nameForAddress: address timeout: seconds! !

!DnsClient class methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:44'!
reset
	"Reset the default DnsClient"

	Default := nil.
! !

!DnsClient class methodsFor: 'class initialization' stamp: 'ar 7/26/2010 22:44'!
shutDown
	"Reset DnsClient to forget all cached information in the snapshot"

	self reset.
! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/26/2010 21:07'!
addressForName: hostName
	"Return an address for the given host name"

	^self addressForName: hostName timeout: 15! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/26/2010 21:07'!
addressForName: hostName timeout: seconds
	"Return an address for the given host name"

	^(self addressesForName: hostName timeout: seconds)
		ifEmpty:[^(NameLookupFailure hostName: hostName) 
					signal: 'Could not resolve: ', hostName]
		ifNotEmpty:[:list| list atRandom]! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/26/2010 21:35'!
addressesForName: hostName
	"Return all the (IPv4) addresses for the given hostName"

	^self addressesForName: hostName timeout: 15! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/27/2010 00:14'!
addressesForName: hostName timeout: seconds
	"Return all the (IPv4) addresses for the given hostName"

	| entry canonical |
	"Canonicalize the host name"
	canonical := hostName asLowercase.
	canonical = 'localhost' ifTrue:[^Array with: #(127 0 0 1) asByteArray].

	"Look up the host name in the cache"
	entry := self cacheAt: canonical.
	((entry at: 1)  >= Time totalSeconds) ifTrue:[^entry at: 2].

	"If no cache hit, go look it up from other DNS servers"
	self lookup: canonical deadline: (Time totalSeconds + seconds).

	"Return the response even if it is expired if we haven't been able to look it up again"
	entry := self cacheAt: canonical.
	^entry at: 2! !

!DnsClient methodsFor: 'accessing' stamp: 'ar 7/26/2010 21:47'!
autoUpdate
	"If true, updates the server list every five minutes."

	^nextUpdate notNil! !

!DnsClient methodsFor: 'accessing' stamp: 'ar 7/26/2010 21:47'!
autoUpdate: aBool
	"If true, updates the server list every five minutes."

	nextUpdate := aBool ifTrue:[0] ifFalse:[nil].
! !

!DnsClient methodsFor: 'caching' stamp: 'ar 7/26/2010 20:45'!
cacheAt: hostName
	"Answer the cache entry corresponding to the given host name.
	The cache entry is a two element {deadline. addressList} array."

	^self cacheAt: hostName ifAbsent:[#(0 #())].! !

!DnsClient methodsFor: 'caching' stamp: 'ar 7/26/2010 20:45'!
cacheAt: hostName ifAbsent: aBlock
	"Answer the cache entry corresponding to the given host name.
	If the hostName is not in the cache evaluate the argument block."

	^mutex critical:[cache at: hostName ifAbsent:[aBlock value]].! !

!DnsClient methodsFor: 'caching' stamp: 'ar 7/26/2010 20:45'!
cacheAt: hostName put: cacheEntry
	"Store the cache entry corresponding to the given host name..
	The cache entry is a two element {deadline. addressList} array."

	^mutex critical:[cache at: hostName put: cacheEntry].! !

!DnsClient methodsFor: 'initialize' stamp: 'ar 7/26/2010 20:44'!
flushCache
	"Flush the entire DNS cache"

	mutex critical:[cache := Dictionary new].
! !

!DnsClient methodsFor: 'initialize' stamp: 'ar 7/26/2010 21:24'!
initialize
	"Initialize the receiver"

	nextUpdate := 0.
	servers := #().
	cache := Dictionary new.
	mutex := Semaphore forMutualExclusion.
! !

!DnsClient methodsFor: 'private' stamp: 'ar 7/26/2010 21:47'!
lookup: hostName deadline: deadline
	"Look up the given host name"

	| records ttl addresses |
	self updateServerList.

	records := self lookupRecordsFor: hostName deadline: deadline type: 1. " 'A' record"
	records ifEmpty:[^self].

	"Collect any A-records from the result"
	ttl := 999999999.
	addresses := records 
		select:[:dnsRec| dnsRec queryType = 1]
		thenCollect:[:dnsRec| 
			ttl := ttl min: dnsRec ttl.
			dnsRec data
		].

	addresses ifEmpty:[
		"No addresses returned; find a CNAME and look it up instead"
		records do:[:dnsRec| 
			dnsRec queryType = 5 
				ifTrue:[^self lookup: dnsRec data deadline: deadline]].
	] ifNotEmpty:[
		cache at: hostName put: {Time totalSeconds + ttl. addresses}.
	].! !

!DnsClient methodsFor: 'private' stamp: 'ar 7/27/2010 00:30'!
lookupRecordsFor: hostName deadline: deadline type: dnsType
	"Look up the given host name"

	| request data sz bytes result |

	request := DnsMessage new.
	request recursionDesired: true.
	request addQuery: hostName type: dnsType. 

	sz := request byteSize.
	data := ByteArray new: sz + 2 streamContents:[:s|
		s nextNumber: 2 put: sz.
		request writeOn: s.
	].
	data size = (sz+2) ifFalse:[self error: 'Incorrect length'].


	servers do:[:dnsServer|
		bytes := [self sendDnsRequest: data to: dnsServer] 
			valueWithin: 2 seconds  onTimeout:[nil].

		"FIXME. If we have a non-responding DNS server we should push it
		back to the end of the servers list to avoid its use in future lookups. 
		Unfortunately, this makes synchronization of the servers a pain 
		since multiple lookups may happen at the same time."

		bytes ifNotNil:[
			"Check the response. If the response is non-empty and includes 
			any queryType or CNAME records, return those."
			result := DnsMessage readFrom: bytes readStream.
			result responseCode = 0 ifTrue:[
				(result responses 
					select:[:resp| resp queryType =  5 or:[resp queryType = dnsType]])
						ifNotEmpty:[:dnsRecords| ^dnsRecords]]]].
	"No responses, return empty-handed"
	^#()! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/26/2010 20:54'!
mxRecordsFor: hostName
	"Return the MX records for the given host name. NOT cached."

	^self mxRecordsFor: hostName timeout: 15! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/26/2010 21:47'!
mxRecordsFor: hostName timeout: seconds
	"Return the MX records for the given host name. NOT cached."

	| records deadline |
	deadline := Time totalSeconds + seconds.
	self updateServerList.

	records := self lookupRecordsFor: hostName deadline: deadline type: 15. " 'MX' record"
	records ifEmpty:[^#()].

	"Collect any MX-records from the result"
	^records 
		select:[:dnsRec| dnsRec queryType = 15]
		thenCollect:[:dnsRec| dnsRec data at: #mx].
! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/26/2010 21:34'!
nameForAddress: address
	"Perform a reverse DNS lookup for the given address"

	^self nameForAddress: address timeout: 15! !

!DnsClient methodsFor: 'resolving' stamp: 'ar 7/27/2010 00:15'!
nameForAddress: address timeout: seconds
	"Perform a reverse DNS lookup for the given address"

	| entry canonical |
	"Canonicalize the host name"
	canonical := String streamContents:[:s|
		address reversed do:[:byte| s print: byte] separatedBy:[s nextPut: $.].
		s nextPutAll: '.IN-ADDR.ARPA'.
	].

	canonical = '1.0.0.127.IN-ADDR.ARPA' ifTrue:[^'localhost'].

	"Look up the host name in the cache"
	entry := self cacheAt: canonical.
	((entry at: 1)  >= Time totalSeconds) ifTrue:[^entry at: 2].

	"If no cache hit, go look it up from other DNS servers"
	self reverseLookup: canonical deadline: (Time totalSeconds + seconds).

	"Return the response even if it is expired if we haven't been able to look it up again"
	entry := self cacheAt: canonical.
	^(entry at: 2) ifEmpty:[nil] ifNotEmpty:[:list| list first]! !

!DnsClient methodsFor: 'private' stamp: 'ar 7/26/2010 20:38'!
resolvDotConf
	"Private. Answer the contents of /etc/resolv.conf. This is a primitive so that platforms without /etc/resolv.conf can implement the proper means to return appropriate data."

	| file |
	"<primitive: 'primitiveResolvDotConf'>"

	file := [FileStream readOnlyFileNamed: '/etc/resolv.conf'] on: Error do:[nil].
	file ifNotNil:[^file contentsOfEntireFile].

	"Use a standard resolv.conf file, made up of Google public DNS and OpenDNS"
	^'# Default resolv.conf file for DnsClient
nameserver 8.8.8.8				# Google Public DNS server 1
nameserver 8.8.4.4				# Google Public DNS server 2
nameserver 208.67.222.222	# OpenDNS server 1
nameserver 208.67.220.220	# OpenDNS server 2
'! !

!DnsClient methodsFor: 'private' stamp: 'ar 7/26/2010 21:48'!
reverseLookup: hostName deadline: deadline
	"Look up the given host name"

	| records ttl addresses |
	self updateServerList.

	records := self lookupRecordsFor: hostName deadline: deadline type: 12. " 'PTR' record"

	"Collect any PTR-records from the result"
	ttl := 999999999.
	addresses := records 
		select:[:dnsRec| dnsRec queryType = 12]
		thenCollect:[:dnsRec| 
			ttl := ttl min: dnsRec ttl.
			dnsRec data
		].

	addresses ifEmpty:[
		"No addresses returned; find a CNAME and look it up instead"
		records do:[:dnsRec| 
			dnsRec queryType = 5 
				ifTrue:[^self lookup: dnsRec data deadline: deadline]].
	] ifNotEmpty:[
		cache at: hostName put: {Time totalSeconds + ttl. addresses}.
	].! !

!DnsClient methodsFor: 'private' stamp: 'ar 7/27/2010 00:03'!
sendDnsRequest: data to: dnsServer
	"Send a DNS request on the given socket and return the response"

	| tcpProcess tcpSocket tcpBytes udpProcess udpSocket udpBytes sema |
	[sema := Semaphore new.
	tcpSocket := Socket newTCP.
	tcpProcess := [
		[tcpSocket connectTo: dnsServer port: 53.
		tcpSocket waitForConnectionFor: 2.
		tcpSocket isConnected ifTrue:[
			tcpSocket sendData: data.
			tcpSocket waitForDataFor: 2.
			tcpSocket dataAvailable ifTrue:[ | count |
				count := ByteArray new: 2.
				(tcpSocket receiveDataInto: count) = 2 ifTrue:[
					tcpBytes := ByteArray new: (count readStream nextNumber: 2).
					(tcpSocket receiveDataInto: tcpBytes) = tcpBytes size
						ifFalse:[tcpBytes := nil]]]]] on: Error do:[].
		sema signal.
	] fork.
	udpSocket := Socket newUDP.
	udpProcess := [[ | count tmpBytes |
		tmpBytes := ByteArray new: 512.
		udpSocket setPeer: dnsServer port: 53.
		udpSocket sendSomeData: data startIndex: 3.
		count := udpSocket receiveDataInto: tmpBytes.
		count > 0 ifTrue:[udpBytes := tmpBytes copyFrom: 1 to: count]] on: Error do:[].
		sema signal.
	] fork.
	(sema waitTimeoutSeconds: 2) ifFalse:[
		(tcpBytes == nil and:[udpBytes == nil]) 
			ifTrue:[sema waitTimeoutSeconds: 2]].
	] ensure:[
		tcpProcess ifNotNil:[tcpProcess terminate].
		udpProcess ifNotNil:[udpProcess terminate].
		tcpSocket ifNotNil:[tcpSocket destroy].
		udpSocket ifNotNil:[udpSocket destroy].
	].
	^tcpBytes ifNil:[udpBytes]! !

!DnsClient methodsFor: 'accessing' stamp: 'ar 7/26/2010 21:46'!
servers
	"The list of DNS servers currently in use.
	If the auto-update property is set, the list of servers can change at any time."

	^servers! !

!DnsClient methodsFor: 'accessing' stamp: 'ar 7/26/2010 21:46'!
servers: anArray
	"The list of DNS servers currently in use
	If the auto-update property is set, the list of servers can change at any time."

	servers := anArray.! !

!DnsClient methodsFor: 'private' stamp: 'ar 7/26/2010 21:50'!
updateServerList
	"Private. Update the list of servers from /etc/resolv.conf.

	This method is deliberately not synchronized; the only access to DnsClient
	are the stores at the end (which are atomic)."

	| data result |

	(self autoUpdate and:[Time totalSeconds > nextUpdate]) ifFalse:[^self].

	result := OrderedCollection new.
	data := self resolvDotConf.
	data linesDo:[:string|
		(string beginsWith: 'nameserver ') ifTrue:[
			(((string copyAfter: Character space) copyUpTo: $#) findTokens: ' ,') do:[:dns|
				result add: ((dns splitBy: '.') collect:[:each| each asNumber] as: ByteArray).
			].
		].
	].
	servers := result asArray.
	nextUpdate := Time totalSeconds + 300.
! !

Object subclass: #DnsMessage
	instanceVariableNames: 'id flags queries responses authorities extras'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'DnsClient-Core'!

!DnsMessage class methodsFor: 'instance creation' stamp: 'ar 7/26/2010 22:18'!
readFrom: aStream
	"Read a DnsMessage from the given stream"

	^self new readFrom: aStream! !

!DnsMessage methodsFor: 'queries' stamp: 'ar 7/26/2010 22:20'!
addQuery: hostName
	"Add a query for the given hostName and ANY queryType"

	^self addQuery: hostName type: 255! !

!DnsMessage methodsFor: 'queries' stamp: 'ar 7/26/2010 22:20'!
addQuery: hostName type: queryType
	"Add a query for the given hostName, queryType and IN query class."

	^self addQuery: hostName type: queryType class: 1 "IN"! !

!DnsMessage methodsFor: 'queries' stamp: 'ar 7/26/2010 22:20'!
addQuery: hostName type: queryType class: queryClass
	"Add a query for the given hostName, queryType and queryClass."

	queries := queries copyWith: (DnsQuery name: hostName type: queryType class: queryClass)! !

!DnsMessage methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:20'!
asByteArray
	"Convert the receiver into a ByteArray"

	^ByteArray streamContents:[:s| self writeOn: s].
! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
authoritative
	"Flag bit indicating that the server creating the response is authoritative"

	^flags anyMask: 1024! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
authoritative: aBool
	"Flag bit indicating that the server creating the response is authoritative"

	flags := aBool ifTrue:[flags bitOr: 1024] ifFalse:[flags bitClear: 1024]! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
authorities
	"The name servers contained in this message"

	^authorities! !

!DnsMessage methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:20'!
byteSize
	"Answer the number of bytes this message will take"

	^12 "id+flags+sizes"
		+ (queries inject: 0 into:[:sz :rec| sz + rec byteSize])
		+ (responses inject: 0 into:[:sz :rec| sz + rec byteSize])
		+ (authorities inject: 0 into:[:sz :rec| sz + rec byteSize])
		+ (extras inject: 0 into:[:sz :rec| sz + rec byteSize])! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
extras
	"The additional records contained in this message"

	^extras! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
id
	"A 16 bit id field generated by the client creating the DNS request.
	Used to match up server responses with the outstanding requests."

	^id! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
id: aNumber
	"A 16 bit id field generated by the client creating the DNS request.
	Used to match up server responses with the outstanding requests."

	id := aNumber! !

!DnsMessage methodsFor: 'initialize' stamp: 'ar 7/27/2010 00:31'!
initialize
	"Initialize the receiver"

	id := 16rFFFF atRandom.
	flags := 0.
	queries := responses := authorities := extras := #().! !

!DnsMessage methodsFor: 'testing' stamp: 'ar 7/26/2010 22:19'!
isQuery
	"Returns true if this is a DNS query message"

	^self isResponse not! !

!DnsMessage methodsFor: 'testing' stamp: 'ar 7/26/2010 22:19'!
isResponse
	"Returns true if this is a DNS response message"

	^self responseFlag! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:18'!
opcode
	"4 flag bits indicating the type of operation:
		0 - Standard query.
		1 - Inverse query; obsoleted by RFC 3425
		2 - Server status request
		3 - Reserved.
		4 - Notify message defined by RFC 1996.
		5 - Update message defined by RFC 2136 to implement DynDNS.
	"
	^(flags bitShift: -11) bitAnd: 15! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
opcode: aNumber
	"4 flag bits indicating the type of operation:
		0 - Standard query.
		1 - Inverse query; obsoleted by RFC 3425
		2 - Server status request
		3 - Reserved.
		4 - Notify message defined by RFC 1996.
		5 - Update message defined by RFC 2136 to implement DynDNS.
	"
	flags := (flags bitClear: 16r7800) bitOr: ((aNumber bitAnd: 15) bitShift: 11)! !

!DnsMessage methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:20'!
printFlagsOn: aStream
	"Print a description of the various flags"

	self isQuery ifTrue:[
		aStream nextPutAll: ', query, code: ', self opcode.
	] ifFalse:[
		aStream nextPutAll: ', response, code: ', self responseCode.
	].
	self authoritative ifTrue:[aStream nextPutAll: ', authoritative'].
	self truncated ifTrue:[aStream nextPutAll: ', truncated'].
	self recursionDesired ifTrue:[aStream nextPutAll: ', want recursion'].
	self recursionAvailable ifTrue:[aStream nextPutAll: ', recursion avail.'].! !

!DnsMessage methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:20'!
printOn: aStream
	"Print the receiver on aStream"

	aStream nextPutAll: self class name.
	aStream nextPut: $(.
	aStream nextPutAll: 'id: ', id.
	self printFlagsOn: aStream.
	aStream nextPutAll: ')'.
! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
queries
	"The questions contained in this message"

	^queries! !

!DnsMessage methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:20'!
readFrom: aStream
	"Read a DNS response from the given (binary) stream"

	| nQueries nResponses nAuthorities nExtras|
	id := aStream nextNumber: 2.
	flags := aStream nextNumber: 2.

	nQueries := aStream nextNumber: 2.
	nResponses := aStream nextNumber: 2.
	nAuthorities := aStream nextNumber: 2.
	nExtras := aStream nextNumber: 2.

	queries := (1 to: nQueries) collect:[:i| DnsQuery readFrom: aStream].
	responses := (1 to: nResponses) collect:[:i| DnsRecord readFrom: aStream].
	authorities := (1 to: nAuthorities) collect:[:i| DnsRecord readFrom: aStream].
	extras := (1 to: nExtras) collect:[:i| DnsRecord readFrom: aStream].! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
recursionAvailable
	"Flag bit indicating that the server can resolve recursively"

	^flags anyMask: 128! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
recursionAvailable: aBool
	"Flag bit indicating that the server can resolve recursively"

	flags := aBool ifTrue:[flags bitOr: 128] ifFalse:[flags bitClear: 128]! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
recursionDesired
	"Flag bit indicating that the server should resolve recursively"

	^flags anyMask: 256! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
recursionDesired: aBool
	"Flag bit indicating that the server should resolve recursively"

	flags := aBool ifTrue:[flags bitOr: 256] ifFalse:[flags bitClear: 256]
! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
responseCode
	"4 bits indicating the server response code:
		0 - No error.
		1 - Format error.
		2 - Server failure.
		3 - Name error.
		4 - Not implemented.
		5 - Refused.
		6 - YX Domain.
		7 - YX RR Set.
		8 - NX RR Set.
		9 - Not Auth. The server is not authoritative for the specified zone.
		10 - Not Zone. A name in the message is not within the zone specified in the message.
	"
	^flags bitAnd: 15! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
responseCode: aNumber
	"4 bits indicating the server response code:
		0 - No error.
		1 - Format error.
		2 - Server failure.
		3 - Name error.
		4 - Not implemented.
		5 - Refused.
		6 - YX Domain.
		7 - YX RR Set.
		8 - NX RR Set.
		9 - Not Auth. The server is not authoritative for the specified zone.
		10 - Not Zone. A name in the message is not within the zone specified in the message.
	"
	flags := (flags bitClear: 15) bitOr: (aNumber bitAnd: 15)! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
responseFlag
	"Flag bit indicating that this is a respone a previous query"

	^flags anyMask: 16r8000! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
responseFlag: aBool
	"Flag bit indicating that this is a respone a previous query"

	flags := aBool ifTrue:[flags bitOr: 16r8000] ifFalse:[flags bitClear: 16r8000].! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
responses
	"The answers contained in this message"

	^responses! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
truncated
	"Flag bit indicating that the message was truncated"

	^flags anyMask: 512! !

!DnsMessage methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:19'!
truncated: aBool
	"Flag bit indicating that the message was truncated"

	flags := aBool ifTrue:[flags bitOr: 512] ifFalse:[flags bitClear: 512]! !

!DnsMessage methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:21'!
writeOn: aStream
	"Store the DNS message on the given (binary) stream"

	aStream nextNumber: 2 put: id.
	aStream nextNumber: 2 put: flags.

	aStream nextNumber: 2 put: queries size.
	aStream nextNumber: 2 put: responses size.
	aStream nextNumber: 2 put: authorities size.
	aStream nextNumber: 2 put: extras size.

	queries do:[:each| each writeOn: aStream].
	responses do:[:each| each writeOn: aStream].
	authorities do:[:each| each writeOn: aStream].
	extras do:[:each| each writeOn: aStream].! !

Object subclass: #DnsQuery
	instanceVariableNames: 'name queryType queryClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'DnsClient-Core'!

!DnsQuery class methodsFor: 'instance creation' stamp: 'ar 7/26/2010 22:23'!
name: hostName type: typeId class: classId
	"Create a new instance of the receiver with the given attributes"
	
	^self new name: hostName type: typeId class: classId! !

!DnsQuery class methodsFor: 'instance creation' stamp: 'ar 7/26/2010 22:23'!
readFrom: aStream
	"Reads a new instance of the receiver from the stream"

	^self new readFrom: aStream! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 20:30'!
byteSize
	"Answer the number of bytes this query takes"

	^name size + 6! !

!DnsQuery methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:21'!
name
	"The name involved in the DNS query"

	^name! !

!DnsQuery methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:21'!
name: aString
	"The name involved in the DNS query"

	name := aString! !

!DnsQuery methodsFor: 'initialize' stamp: 'ar 7/26/2010 22:21'!
name: hostName type: typeId class: classId
	"Initialize the receiver using name, type, and class id"

	name := hostName.
	queryType := typeId.
	queryClass := classId.! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:21'!
nameForRecord: typeId
	"Answer the printable name of a given DNS record type"

	typeId = 255 ifTrue:[^'*'].
	typeId = 28 ifTrue:[^'AAAA'].
	(typeId between: 1 and: 15) ifFalse:[^typeId printString].
	^#(A NS MD MF CNAME SOA MB MG MR NULL WKS PTR HINFO MINFO MX) at: typeId! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:22'!
printExtrasOn: aStream
	"Print extra information on aStream"
! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:22'!
printOn: aStream
	"Print the receiver on aStream"

	aStream nextPutAll: self class name.
	aStream nextPut: $(.
	aStream nextPutAll: name.
	aStream nextPutAll: ', type: '.
	aStream nextPutAll: (self nameForRecord: queryType).
	self printExtrasOn: aStream.
	aStream nextPutAll: ')'.
! !

!DnsQuery methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:21'!
queryClass
	"The class of query to perform (1 == INternet queries)"

	^queryClass! !

!DnsQuery methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:21'!
queryClass: classId
	"The class of query to perform (1 == INternet queries)"

	queryClass := classId! !

!DnsQuery methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:21'!
queryType
	"The type of query being performed"

	^queryType! !

!DnsQuery methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:21'!
queryType: typeId
	"The type of query being performed"

	queryType := typeId.! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:22'!
readFrom: aStream
	"Read the query from the given stream"

	name := self readNameFrom: aStream.
	queryType := aStream nextNumber: 2.
	queryClass := aStream nextNumber: 2.
! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:22'!
readNameFrom: aStream
	"Reads an encoded name from the given stream"

	^String streamContents:[:s| | count |
		[count := aStream next.
		count = 0] whileFalse:[
			s position = 0 ifFalse:[s nextPut: $.].
			count >= 192 ifTrue:[ | offset endpos |
				offset := (count bitAnd:  63) * 256 + aStream next.
				endpos := aStream position.
				aStream position: offset.
				s nextPutAll: (self readNameFrom: aStream).
				aStream position: endpos.
				^s contents
			] ifFalse:[s nextPutAll: (aStream next: count) asString].
		].
	].
! !

!DnsQuery methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:22'!
writeOn: aStream
	"Writes the query on the given stream"

	(name splitBy: '.') do:[:label|
		(label size between: 1 and: 63) ifFalse:[self error: 'Invalid host name'].
		aStream nextPut: label size.
		"XXX: Fixme. What encoding?"
		aStream nextPutAll: label asByteArray.
	].
	aStream nextPut: 0.
	aStream nextNumber: 2 put: queryType.
	aStream nextNumber: 2 put: queryClass.
! !

DnsQuery subclass: #DnsRecord
	instanceVariableNames: 'ttl data'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'DnsClient-Core'!

!DnsRecord methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:23'!
byteSize
	"Answer the number of bytes this record takes"

	^super byteSize + 6 + data size! !

!DnsRecord methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:23'!
data
	"The record data"

	^data! !

!DnsRecord methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:23'!
data: anObject
	"The record data"

	data := anObject! !

!DnsRecord methodsFor: 'initialize' stamp: 'ar 7/26/2010 22:23'!
initialize
	"Initialize the receiver"

	super initialize.
	ttl := 0.
	data := ByteArray new.! !

!DnsRecord methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:23'!
printExtrasOn: aStream
	"Prints extra information for this record"

	(data isKindOf: Dictionary)
		ifTrue:[^data keysAndValuesDo:[:key :val| aStream nextPutAll: ', ', key, ': '; print: val]].
	aStream nextPutAll:', '; print: data.! !

!DnsRecord methodsFor: 'read/write' stamp: 'fbs 7/27/2010 10:12'!
readDataFrom: aStream
	"Reads and decodes the response specific data"

	| n |
	n := aStream nextNumber: 2.
	^queryType caseOf: {
		[2] 		-> [self readNameFrom: aStream].	"NS"
		[5] 		-> [self readNameFrom: aStream].	"CNAME"
		[12]	-> [self readNameFrom: aStream]. 	"PTR"
		[15]	-> [Dictionary newFromPairs: {		"MX"
						#priority. 	aStream nextNumber: 2.
						#mx. 		self readNameFrom: aStream. 
					}].
		[33]	-> [Dictionary newFromPairs: {		"SRV"
						#priority.	aStream nextNumber: 2.
						#weight.	aStream nextNumber: 2.
						#port.		aStream nextNumber: 2.
						#target.	self readNameFrom: aStream.
					}].
		[35]	-> [Dictionary newFromPairs: {		"NAPTR"
						#order.			aStream nextNumber: 2.
						#preference.	aStream nextNumber: 2.
						#flags.			aStream nextNumber: 2.
						#services.		aStream nextNumber: 2.
						#regexp.		aStream nextNumber: 2.
						#replacement.	aStream nextNumber: 2.
					}].
		[39] 	-> [self readNameFrom: aStream]. 	"DNAME"
	} otherwise:[aStream next: n].
! !

!DnsRecord methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:24'!
readFrom: aStream
	"Read a DNS record from the given stream"

	super readFrom: aStream.
	ttl := aStream nextNumber: 4.
	data := self readDataFrom: aStream.
! !

!DnsRecord methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:23'!
ttl
	"The time-to-live for this record"

	^ttl! !

!DnsRecord methodsFor: 'accessing' stamp: 'ar 7/26/2010 22:23'!
ttl: seconds
	"The time-to-live for this record"

	ttl := seconds! !

!DnsRecord methodsFor: 'read/write' stamp: 'ar 7/26/2010 22:24'!
writeOn: aStream
	"Store a DNS record from the given stream"

	super writeOn: aStream.
	aStream nextNumber: 4 put: ttl.
	aStream nextNumber: 2 put: data size.
	aStream nextPutAll: data.! !
DnsClient initialize!
